@@ -109,6 +109,16 @@
 #define BIOS_VIDEO_INFO_2               0x489
 #define BIOS_VIDEO_COMBO                0x48a
 
+#define BIOS_KEYBOARD_PCJR_FLAG2        0x488 /* KB_FLAG_2 */
+#define BIOS_KEYBOARD_PCJR_FLAG2_PUTCHAR     (1 << 0)
+#define BIOS_KEYBOARD_PCJR_FLAG2_INIT_DELAY  (1 << 1)
+#define BIOS_KEYBOARD_PCJR_FLAG2_HALF_RATE   (1 << 2)
+#define BIOS_KEYBOARD_PCJR_FLAG2_TYPE_OFF    (1 << 3)
+#define BIOS_KEYBOARD_PCJR_FLAG2_FN_LOCK     (1 << 4)
+#define BIOS_KEYBOARD_PCJR_FLAG2_FN_PENDING  (1 << 5)
+#define BIOS_KEYBOARD_PCJR_FLAG2_FN_BREAK    (1 << 6)
+#define BIOS_KEYBOARD_PCJR_FLAG2_FN_FLAG     (1 << 7)
+
 #define BIOS_KEYBOARD_FLAGS3            0x496
 #define BIOS_KEYBOARD_FLAGS3_HIDDEN_E1			(1 << 0)
 #define BIOS_KEYBOARD_FLAGS3_HIDDEN_E0			(1 << 1)
@@ -164,7 +174,7 @@ Bitu ROMBIOS_GetMemory(Bitu bytes,const char *who=NULL,Bitu alignment=1,Bitu mus
 extern RegionAllocTracking rombios_alloc;
 
 /* maximum of scancodes handled by keyboard bios routines */
-#define MAX_SCAN_CODE 0x93
+#define MAX_SCAN_CODE 0x7F
 
 /* The Section handling Bios Disk Access */
 //#define BIOS_MAX_DISK 10
@@ -179,6 +189,13 @@ void INT10_ReloadRomFonts();
 
 void BIOS_SetLPTPort (Bitu port, uint16_t baseaddr);
 
+// Is it safe to call mem_read/write as if part of the guest i.e. from the main GUI?
+// If the guest is running in protected mode, NO.
+// If the guest is not running under our own DOS kernel, NO.
+// This is required to avoid problems with helpful code in this project causing BSODs in Windows NT/2000/XP
+// because the BIOS data area addresses are invalid pages in that 32-bit environment!
+bool IsSafeToMemIOOnBehalfOfGuest();
+
 // \brief Synchronizes emulator num lock state with host.
 void BIOS_SynchronizeNumLock();
 
