@@ -26,6 +26,11 @@
 #include "cpu.h"
 #include "logging.h"
 
+extern bool do_pse;
+extern bool enable_pse;
+extern uint8_t enable_pse_extbits;
+extern uint8_t enable_pse_extmask;
+
 extern bool dos_kernel_disabled;
 PagingBlock paging;
 
@@ -113,7 +118,7 @@ Bits PageFaultCore(void) {
 	if (ret) 
 		return ret;
 	if (!pf_queue.used) E_Exit("PF Core without PF");
-    const PF_Entry* entry = &pf_queue.entries[pf_queue.used - 1];
+	const PF_Entry* entry = &pf_queue.entries[pf_queue.used - 1];
 	X86PageEntry pentry;
 	pentry.load=phys_readd((PhysPt)entry->page_addr);
 	if (pentry.block.p && entry->cs == SegValue(cs) && entry->eip==reg_eip) {
@@ -216,16 +221,14 @@ static const uint8_t fault_table[] = {
 	1,	1,	1,	0,
 };
 
-#define PHYSPAGE_DITRY 0x10000000
-#define PHYSPAGE_ADDR  0x000FFFFF
-
 // helper functions for calculating table entry addresses
 static inline PhysPt GetPageDirectoryEntryAddr(PhysPt lin_addr) {
 	return paging.base.addr | ((lin_addr >> 22u) << 2u);
 }
 static inline PhysPt GetPageTableEntryAddr(PhysPt lin_addr, const X86PageEntry& dir_entry) {
-	return ((PhysPt)dir_entry.block.base << (PhysPt)12U) | ((lin_addr >> 10U) & 0xffcu);
+	return ((PhysPt)dir_entry.dirblock.base << (PhysPt)12U) | ((lin_addr >> 10U) & 0xffcu);
 }
+
 /*
 void PrintPageInfo(const char* string, PhysPt lin_addr, bool writing, bool prepare_only) {
 
@@ -241,7 +244,7 @@ void PrintPageInfo(const char* string, PhysPt lin_addr, bool writing, bool prepa
 	bool dirty = false;
 	Bitu ft_index = 0;
 
-	if (dir_entry.block.p) {
+	if (dir_entry.dirblock.p) {
 		tableEntryAddr = GetPageTableEntryAddr(lin_addr, dir_entry);
 		table_entry.load=phys_readd(tableEntryAddr);
 		if (table_entry.block.p) {
@@ -341,45 +344,55 @@ private:
 		uint32_t phys_page = paging.tlb.phys_page[lin_page] & PHYSPAGE_ADDR;
 			
 		// set the page dirty in the tlb
-		paging.tlb.phys_page[lin_page] |= PHYSPAGE_DITRY;
+		paging.tlb.phys_page[lin_page] |= PHYSPAGE_DIRTY;
 
 		// mark the page table entry dirty
 		X86PageEntry dir_entry, table_entry;
 
 		PhysPt dirEntryAddr = GetPageDirectoryEntryAddr(addr);
-		dir_entry.load=phys_readd(dirEntryAddr);
-		if (!dir_entry.block.p) E_Exit("Undesired situation 1 in page foiler.");
+		dir_entry.load = phys_readd(dirEntryAddr);
+		if (!dir_entry.dirblock.p) E_Exit("Undesired situation 1 in page foiler.");
 
-		PhysPt tableEntryAddr = GetPageTableEntryAddr(addr, dir_entry);
-		table_entry.load=phys_readd(tableEntryAddr);
-		if (!table_entry.block.p) E_Exit("Undesired situation 2 in page foiler.");
+		if (do_pse && dir_entry.dirblock.ps) {
+			// for debugging...
+			if (((dir_entry.dirblock4mb.base22<<10u)|((dir_entry.dirblock4mb.base32&enable_pse_extmask)<<20u)|(lin_page&0x3FFu)) != phys_page)
+				E_Exit("Undesired situation 3 PSE in page foiler.");
 
-		// for debugging...
-		if (table_entry.block.base != phys_page)
-			E_Exit("Undesired situation 3 in page foiler.");
+			// set the dirty bit
+			dir_entry.block.d=1;
+			phys_writed(dirEntryAddr,dir_entry.load);
+		}
+		else {
+			PhysPt tableEntryAddr = GetPageTableEntryAddr(addr, dir_entry);
+			table_entry.load=phys_readd(tableEntryAddr);
+			if (!table_entry.block.p) E_Exit("Undesired situation 2 in page foiler.");
+
+			// for debugging...
+			if (table_entry.block.base != phys_page)
+				E_Exit("Undesired situation 3 in page foiler.");
+
+			// debug
+//			LOG_MSG("FOIL            LIN% 8x PHYS% 8x [%x/%x/%x] WRP % 8x", addr, phys_page,
+//				dirEntryAddr, tableEntryAddr, table_entry.load, wtest);
+
+			// this can happen when the same page table is used at two different
+			// page directory entries / linear locations (WfW311)
+			// if (table_entry.block.d) E_Exit("% 8x Already dirty!!",table_entry.load);
+
+			// set the dirty bit
+			table_entry.block.d=1;
+			phys_writed(tableEntryAddr,table_entry.load);
+		}
 
-		// map the real write handler in our place
 		PageHandler* handler = MEM_GetPageHandler(phys_page);
 
-		// debug
-//		LOG_MSG("FOIL            LIN% 8x PHYS% 8x [%x/%x/%x] WRP % 8x", addr, phys_page,
-//			dirEntryAddr, tableEntryAddr, table_entry.load, wtest);
-
-		// this can happen when the same page table is used at two different
-		// page directory entries / linear locations (WfW311)
-		// if (table_entry.block.d) E_Exit("% 8x Already dirty!!",table_entry.load);
-		
-		
-		// set the dirty bit
-		table_entry.block.d=1;
-		phys_writed(tableEntryAddr,table_entry.load);
-		
 		// replace this handler with the real thing
 		if (handler->getFlags() & PFLAG_WRITEABLE)
 			paging.tlb.write[lin_page] = handler->GetHostWritePt(phys_page) - (lin_page << 12);
-		else paging.tlb.write[lin_page]=nullptr;
-		paging.tlb.writehandler[lin_page]=handler;
+		else
+			paging.tlb.write[lin_page] = nullptr;
 
+		paging.tlb.writehandler[lin_page]=handler;
 		return;
 	}
 
@@ -436,28 +449,41 @@ private:
 		uint32_t phys_page = paging.tlb.phys_page[lin_page] & PHYSPAGE_ADDR;
 		PageHandler* handler = MEM_GetPageHandler(phys_page);
 		return handler;
-					}
+	}
 
 	bool hack_check(PhysPt addr) {
 		// First Encounters
+		//
 		// They change the page attributes without clearing the TLB.
 		// On a real 486 they get away with it because its TLB has only 32 or so 
 		// elements. The changed page attribs get overwritten and re-read before
 		// the exception happens. Here we have gazillions of TLB entries so the
 		// exception occurs if we don't check for it.
 
-		Bitu old_attirbs = paging.tlb.phys_page[addr>>12] >> 30;
-		X86PageEntry dir_entry, table_entry;
-		
-		dir_entry.load = phys_readd(GetPageDirectoryEntryAddr(addr));
-		if (!dir_entry.block.p) return false;
-		table_entry.load = phys_readd(GetPageTableEntryAddr(addr, dir_entry));
-		if (!table_entry.block.p) return false;
-		Bitu result =
-		translate_array[((dir_entry.load<<1)&0xc) | ((table_entry.load>>1)&0x3)];
-		if (result != old_attirbs) return true;
+		// NTS 2024/12/15: Checking with 7-cpu.com, the Pentium also has 32,
+		// the Pentium II and III has 64. You would need to go to a pretty late
+		// level of CPU to find something that would probably cause this game to
+		// fail from page table laziness. However, to simplify this code, this
+		// check is skipped over if PSE or PAE are enabled because DOS games are
+		// unlikely to use those page table features. --J.C.
+
+		if (!do_pse) {
+			Bitu old_attirbs = paging.tlb.phys_page[addr>>12] >> 30;
+			X86PageEntry dir_entry, table_entry;
+
+			dir_entry.load = phys_readd(GetPageDirectoryEntryAddr(addr));
+			if (!dir_entry.dirblock.p) return false;
+			table_entry.load = phys_readd(GetPageTableEntryAddr(addr, dir_entry));
+			if (!table_entry.block.p) return false;
+
+			const Bitu result =
+				translate_array[((dir_entry.load<<1)&0xc) | ((table_entry.load>>1)&0x3)];
+
+			if (result != old_attirbs) return true;
+		}
+
 		return false;
-				}
+	}
 
 	void Exception(PhysPt addr, bool writing, bool checked) {
 		//PrintPageInfo("XCEPT",addr,writing, checked);
@@ -466,8 +492,10 @@ private:
 		if (!checked) {
 			X86PageEntry dir_entry;
 			dir_entry.load = phys_readd(GetPageDirectoryEntryAddr(addr));
-			if (!dir_entry.block.p) E_Exit("Undesired situation 1 in exception handler.");
-			
+			if (!dir_entry.dirblock.p) E_Exit("Undesired situation 1 in exception handler.");
+
+			if (do_pse && dir_entry.dirblock.ps) E_Exit("PSE and Exception not yet supported");//TODO
+
 			// page table entry
 			tableaddr = GetPageTableEntryAddr(addr, dir_entry);
 			//Bitu d_index=(addr >> 12) >> 10;
@@ -646,15 +674,21 @@ static INLINE void InitPageCheckPresence(PhysPt lin_addr,bool writing,X86PageEnt
 		if (GCC_UNLIKELY(!table.block.p))
 			E_Exit("Pagefault didn't correct table");
 	}
-	Bitu entry_addr=(table.block.base<<12)+t_index*4;
-	entry.load=phys_readd((PhysPt)entry_addr);
-	if (!entry.block.p) {
-//		LOG(LOG_PAGING,LOG_NORMAL)("NP Page");
-		PAGING_PageFault(lin_addr,entry_addr,
-			(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04));
+
+	if (do_pse && table.dirblock.ps) { // 4MB PSE page
+		entry.load=0;
+	}
+	else {
+		Bitu entry_addr=(table.block.base<<12)+t_index*4;
 		entry.load=phys_readd((PhysPt)entry_addr);
-		if (GCC_UNLIKELY(!entry.block.p))
-			E_Exit("Pagefault didn't correct page");
+		if (!entry.block.p) {
+//			LOG(LOG_PAGING,LOG_NORMAL)("NP Page");
+			PAGING_PageFault(lin_addr,entry_addr,
+				(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04));
+			entry.load=phys_readd((PhysPt)entry_addr);
+			if (GCC_UNLIKELY(!entry.block.p))
+				E_Exit("Pagefault didn't correct page");
+		}
 	}
 }
 
@@ -670,14 +704,20 @@ static INLINE bool InitPageCheckPresence_CheckOnly(PhysPt lin_addr,bool writing,
 		cpu.exception.error=(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04);
 		return false;
 	}
-	Bitu entry_addr=(table.block.base<<12)+t_index*4;
-	entry.load=phys_readd((PhysPt)entry_addr);
-	if (!entry.block.p) {
-		paging.cr2=lin_addr;
-		cpu.exception.which=EXCEPTION_PF;
-		cpu.exception.error=(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04);
-		return false;
+
+	if (do_pse && table.dirblock.ps) { // 4MB PSE page
 	}
+	else {
+		Bitu entry_addr=(table.block.base<<12)+t_index*4;
+		entry.load=phys_readd((PhysPt)entry_addr);
+		if (!entry.block.p) {
+			paging.cr2=lin_addr;
+			cpu.exception.which=EXCEPTION_PF;
+			cpu.exception.error=(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04);
+			return false;
+		}
+	}
+
 	return true;
 }
 
@@ -766,94 +806,120 @@ initpage_retry:
 			PhysPt dirEntryAddr = GetPageDirectoryEntryAddr(lin_addr);
 			// Range check to avoid emulator segfault: phys_readd() reads from MemBase+addr and does NOT range check.
 			// Needed to avoid segfault when running 1999 demo "Void Main" in a bootable Windows 95 image in pure DOS mode.
-			if ((dirEntryAddr+4) <= (MEM_TotalPages()<<12u)) {
-				dir_entry.load=phys_readd(dirEntryAddr);
-			}
-			else {
-				LOG(LOG_CPU,LOG_WARN)("Page directory access beyond end of memory, page %08x >= %08x",(unsigned int)(dirEntryAddr>>12u),(unsigned int)MEM_TotalPages());
-				dir_entry.load=0xFFFFFFFF;
-			}
+			// 2024/12/22: phys_readx() does range checking now
+			dir_entry.load=phys_readd(dirEntryAddr);
 
-			if (!dir_entry.block.p) {
+			if (!dir_entry.dirblock.p) {
 				// table pointer is not present, do a page fault
 				PAGING_NewPageFault(lin_addr, dirEntryAddr, prepare_only,
 					(writing ? 2u : 0u) | (isUser ? 4u : 0u));
-				
+
 				if (prepare_only) return true;
 				else goto initpage_retry; // TODO maybe E_Exit after a few loops
 			}
-			PhysPt tableEntryAddr = GetPageTableEntryAddr(lin_addr, dir_entry);
-			// Range check to avoid emulator segfault: phys_readd() reads from MemBase+addr and does NOT range check.
-			if ((tableEntryAddr+4) <= (MEM_TotalPages()<<12u)) {
-				table_entry.load=phys_readd(tableEntryAddr);
+
+			if (do_pse && dir_entry.dirblock.ps) { // 4MB PSE page
+				Bitu result =
+					translate_array[((dir_entry.load<<1)&0xc) | ((dir_entry.load>>1)&0x3)];
+
+				// save load to see if it changed later
+				uint32_t dir_load = dir_entry.load;
+
+				// if we are writing we can set it right here to save some CPU
+				if (writing) dir_entry.dirblock4mb.d = 1;
+
+				// set page accessed
+				dir_entry.dirblock4mb.a = 1;
+
+				// update if needed
+				if (dir_load != dir_entry.load)
+					phys_writed(dirEntryAddr, dir_entry.load);
+
+				// if the page isn't dirty and we are reading we need to map the foiler
+				// (dirty = false)
+				bool dirty = dir_entry.dirblock4mb.d? true:false;
+
+				/* LOG_MSG("INITPSE lin=0x%x phys=0x%lx base22=0x%x base32=0x%x",
+					(unsigned int)lin_addr,
+					(unsigned long)(((dir_entry.dirblock4mb.base22<<10ul)|((dir_entry.dirblock4mb.base32&enable_pse_extmask)<<20ul)|(lin_page&0x3FFul))<<12ul),
+					(unsigned int)dir_entry.dirblock4mb.base22,
+					(unsigned int)dir_entry.dirblock4mb.base32); */
+				// finally install the new page
+				PAGING_LinkPageNew(lin_page,
+					(dir_entry.dirblock4mb.base22<<10u)|
+					((dir_entry.dirblock4mb.base32&enable_pse_extmask)<<20u)|
+					(lin_page&0x3FFu),
+					result, dirty);
 			}
 			else {
-				LOG(LOG_CPU,LOG_WARN)("Page table entry access beyond end of memory, page %08x >= %08x",(unsigned int)(tableEntryAddr>>12u),(unsigned int)MEM_TotalPages());
-				table_entry.load=0xFFFFFFFF;
+				PhysPt tableEntryAddr = GetPageTableEntryAddr(lin_addr, dir_entry);
+				// Range check to avoid emulator segfault: phys_readd() reads from MemBase+addr and does NOT range check.
+				// 2024/12/22: phys_readx() does range checking now
+				table_entry.load=phys_readd(tableEntryAddr);
+
+				// set page table accessed (IA manual: A is set whenever the entry is 
+				// used in a page translation)
+				if (!dir_entry.dirblock.a) {
+					dir_entry.dirblock.a = 1;
+					phys_writed(dirEntryAddr, dir_entry.load);
+				}
+
+				if (!table_entry.block.p) {
+					// physpage pointer is not present, do a page fault
+					PAGING_NewPageFault(lin_addr, tableEntryAddr, prepare_only,
+						(writing ? 2u : 0u) | (isUser ? 4u : 0u));
+
+					if (prepare_only) return true;
+					else goto initpage_retry;
+				}
+				//PrintPageInfo("INI",lin_addr,writing,prepare_only);
+
+				Bitu result =
+					translate_array[((dir_entry.load<<1)&0xc) | ((table_entry.load>>1)&0x3)];
+
+				// If a page access right exception occurs we shouldn't change a or d
+				// I'd prefer running into the prepared exception handler but we'd need
+				// an additional handler that sets the 'a' bit - idea - foiler read?
+				Bitu ft_index = result | (writing ? 8u : 0u) | (isUser ? 4u : 0u) |
+					(paging.wp ? 16u : 0u);
+
+				if (GCC_UNLIKELY(fault_table[ft_index])) {
+					// exception error code format: 
+					// bit0 - protection violation, bit1 - writing, bit2 - user mode
+					PAGING_NewPageFault(lin_addr, tableEntryAddr, prepare_only,
+						1u | (writing ? 2u : 0u) | (isUser ? 4u : 0u));
+
+					if (prepare_only) return true;
+					else goto initpage_retry; // unlikely to happen?
+				}
+				// save load to see if it changed later
+				uint32_t table_load = table_entry.load;
+
+				// if we are writing we can set it right here to save some CPU
+				if (writing) table_entry.block.d = 1;
+
+				// set page accessed
+				table_entry.block.a = 1;
+
+				// update if needed
+				if (table_load != table_entry.load)
+					phys_writed(tableEntryAddr, table_entry.load);
+
+				// if the page isn't dirty and we are reading we need to map the foiler
+				// (dirty = false)
+				bool dirty = table_entry.block.d? true:false;
+				/*
+				   LOG_MSG("INIT  %s LIN% 8x PHYS% 5x wr%x ch%x wp%x d%x c%x m%x a%x [%x/%x/%x]",
+				   mtr[result], lin_addr, table_entry.block.base,
+				   writing, prepare_only, paging.wp,
+				   dirty, cpu.cpl, cpu.mpl,
+				   ((dir_entry.load<<1)&0xc) | ((table_entry.load>>1)&0x3),
+				   dirEntryAddr, tableEntryAddr, table_entry.load);
+				   */
+				// finally install the new page
+				PAGING_LinkPageNew(lin_page, table_entry.block.base, result, dirty);
 			}
 
-			// set page table accessed (IA manual: A is set whenever the entry is 
-			// used in a page translation)
-			if (!dir_entry.block.a) {
-				dir_entry.block.a = 1;		
-				phys_writed(dirEntryAddr, dir_entry.load);
-		}
-
-			if (!table_entry.block.p) {
-				// physpage pointer is not present, do a page fault
-				PAGING_NewPageFault(lin_addr, tableEntryAddr, prepare_only,
-					 (writing ? 2u : 0u) | (isUser ? 4u : 0u));
-				
-				if (prepare_only) return true;
-				else goto initpage_retry;
-	}
-			//PrintPageInfo("INI",lin_addr,writing,prepare_only);
-
-			Bitu result =
-				translate_array[((dir_entry.load<<1)&0xc) | ((table_entry.load>>1)&0x3)];
-			
-			// If a page access right exception occurs we shouldn't change a or d
-			// I'd prefer running into the prepared exception handler but we'd need
-			// an additional handler that sets the 'a' bit - idea - foiler read?
-			Bitu ft_index = result | (writing ? 8u : 0u) | (isUser ? 4u : 0u) |
-				(paging.wp ? 16u : 0u);
-			
-			if (GCC_UNLIKELY(fault_table[ft_index])) {
-				// exception error code format: 
-				// bit0 - protection violation, bit1 - writing, bit2 - user mode
-				PAGING_NewPageFault(lin_addr, tableEntryAddr, prepare_only,
-					1u | (writing ? 2u : 0u) | (isUser ? 4u : 0u));
-				
-				if (prepare_only) return true;
-				else goto initpage_retry; // unlikely to happen?
-			}
-			// save load to see if it changed later
-			uint32_t table_load = table_entry.load;
-
-			// if we are writing we can set it right here to save some CPU
-			if (writing) table_entry.block.d = 1;
-
-			// set page accessed
-			table_entry.block.a = 1;
-			
-			// update if needed
-			if (table_load != table_entry.load)
-				phys_writed(tableEntryAddr, table_entry.load);
-
-			// if the page isn't dirty and we are reading we need to map the foiler
-			// (dirty = false)
-			bool dirty = table_entry.block.d? true:false;
-/*
-			LOG_MSG("INIT  %s LIN% 8x PHYS% 5x wr%x ch%x wp%x d%x c%x m%x a%x [%x/%x/%x]",
-				mtr[result], lin_addr, table_entry.block.base,
-				writing, prepare_only, paging.wp,
-				dirty, cpu.cpl, cpu.mpl,
-				((dir_entry.load<<1)&0xc) | ((table_entry.load>>1)&0x3),
-				dirEntryAddr, tableEntryAddr, table_entry.load);
-*/
-			// finally install the new page
-			PAGING_LinkPageNew(lin_page, table_entry.block.base, result, dirty);
-
 		} else { // paging off
 			if (lin_page<LINK_START) phys_page=paging.firstmb[lin_page];
 			else phys_page=lin_page;
@@ -873,11 +939,18 @@ initpage_retry:
 				table.block.a=1;		//Set access
 				phys_writed((PhysPt)((paging.base.page<<12)+(lin_page >> 10)*4),table.load);
 			}
-			if (!entry.block.a) {
-				entry.block.a=1;					//Set access
-				phys_writed((table.block.base<<12)+(lin_page & 0x3ff)*4,entry.load);
+			if (do_pse && table.dirblock.ps) { // 4MB PSE page
+				phys_page =	 (table.dirblock4mb.base22<<10u)|
+						((table.dirblock4mb.base32&enable_pse_extmask)<<20u)|
+						 (lin_page&0x3FFu);
+			}
+			else {
+				if (!entry.block.a) {
+					entry.block.a=1;					//Set access
+					phys_writed((table.block.base<<12)+(lin_page & 0x3ff)*4,entry.load);
+				}
+				phys_page=entry.block.base;
 			}
-			phys_page=entry.block.base;
 			// maybe use read-only page here if possible
 		} else {
 			if (lin_page<LINK_START) phys_page=paging.firstmb[lin_page];
@@ -893,15 +966,23 @@ bool PAGING_MakePhysPage(Bitu & page) {
 		// check the page directory entry for this address
 		X86PageEntry dir_entry;
 		dir_entry.load = phys_readd(GetPageDirectoryEntryAddr((PhysPt)(page<<12)));
-		if (!dir_entry.block.p) return false;
-		
-		// check the page table entry
-		X86PageEntry tbl_entry;
-		tbl_entry.load = phys_readd(GetPageTableEntryAddr((PhysPt)(page<<12), dir_entry));
-		if (!tbl_entry.block.p) return false;
+		if (!dir_entry.dirblock.p) return false;
 
-		// return it
-		page = tbl_entry.block.base;
+		if (do_pse && dir_entry.dirblock.ps) {
+			// return it
+			page =	 (dir_entry.dirblock4mb.base22<<10u)|
+				((dir_entry.dirblock4mb.base32&enable_pse_extmask)<<20u)|
+				 (page&0x3FFu);
+		}
+		else {
+			// check the page table entry
+			X86PageEntry tbl_entry;
+			tbl_entry.load = phys_readd(GetPageTableEntryAddr((PhysPt)(page<<12), dir_entry));
+			if (!tbl_entry.block.p) return false;
+
+			// return it
+			page = tbl_entry.block.base;
+		}
 	} else {
 		if (page<LINK_START) page=paging.firstmb[page];
 		//Else keep it the same
@@ -979,9 +1060,12 @@ static void PAGING_LinkPageNew(Bitu lin_page, Bitu phys_page, Bitu linkmode, boo
 	PageHandler * handler=MEM_GetPageHandler(phys_page);
 	Bitu lin_base=lin_page << 12;
 
+	//NTS: phys_page is not used to index anything, however it must not use bits 31-30 used for other info. Stay within PHYSPAGE_ADDR.
+	phys_page &= PHYSPAGE_ADDR;
+
 //	LOG_MSG("MAPPG %s",lnm[outcome]);
 	
-	if (GCC_UNLIKELY(lin_page>=TLB_SIZE || phys_page>=TLB_SIZE)) 
+	if (GCC_UNLIKELY(lin_page>=TLB_SIZE))
 		E_Exit("Illegal page");
 	if (GCC_UNLIKELY(paging.links.used>=PAGING_LINKS)) {
 		LOG(LOG_PAGING,LOG_NORMAL)("Not enough paging links, resetting cache");
@@ -992,21 +1076,21 @@ static void PAGING_LinkPageNew(Bitu lin_page, Bitu phys_page, Bitu linkmode, boo
 	// bit31-30 ACMAP_
 	// bit29	dirty
 	// these bits are shifted off at the places paging.tlb.phys_page is read
-	paging.tlb.phys_page[lin_page]= (uint32_t)(phys_page | (linkmode<< 30) | (dirty? PHYSPAGE_DITRY:0));
+	paging.tlb.phys_page[lin_page]= (uint32_t)(phys_page | (linkmode<< 30) | (dirty? PHYSPAGE_DIRTY:0));
 	switch(outcome) {
 	case ACMAP_RW:
 		// read
 		if (handler->getFlags() & PFLAG_READABLE) paging.tlb.read[lin_page] = 
 			handler->GetHostReadPt(phys_page)-lin_base;
-	else paging.tlb.read[lin_page]=nullptr;
-	paging.tlb.readhandler[lin_page]=handler;
-		
+		else paging.tlb.read[lin_page]=nullptr;
+		paging.tlb.readhandler[lin_page]=handler;
+
 		// write
 		if (dirty) { // in case it is already dirty we don't need to check
 			if (handler->getFlags() & PFLAG_WRITEABLE) paging.tlb.write[lin_page] = 
 				handler->GetHostWritePt(phys_page)-lin_base;
 			else paging.tlb.write[lin_page]=nullptr;
-	paging.tlb.writehandler[lin_page]=handler;
+			paging.tlb.writehandler[lin_page]=handler;
 		} else {
 			paging.tlb.writehandler[lin_page]= &foiling_handler;
 			paging.tlb.write[lin_page]=nullptr;
@@ -1028,21 +1112,21 @@ static void PAGING_LinkPageNew(Bitu lin_page, Bitu phys_page, Bitu linkmode, boo
 		paging.tlb.read[lin_page]=nullptr;
 		paging.tlb.write[lin_page]=nullptr;
 		break;
-}
+	}
 
 	switch(linkmode) {
-	case ACCESS_KR:
-		paging.kr_links.entries[paging.kr_links.used++]=(uint32_t)lin_page;
-		break;
-	case ACCESS_KRW:
-		paging.krw_links.entries[paging.krw_links.used++]= (uint32_t)lin_page;
-		break;
-	case ACCESS_UR:
-		paging.ur_links.entries[paging.ur_links.used++]= (uint32_t)lin_page;
-		break;
-	case ACCESS_URW:	// with this access right everything is possible
-						// thus no need to modify it on a us <-> sv switch
-		break;
+		case ACCESS_KR:
+			paging.kr_links.entries[paging.kr_links.used++]=(uint32_t)lin_page;
+			break;
+		case ACCESS_KRW:
+			paging.krw_links.entries[paging.krw_links.used++]= (uint32_t)lin_page;
+			break;
+		case ACCESS_UR:
+			paging.ur_links.entries[paging.ur_links.used++]= (uint32_t)lin_page;
+			break;
+		case ACCESS_URW:	// with this access right everything is possible
+			// thus no need to modify it on a us <-> sv switch
+			break;
 	}
 	paging.links.entries[paging.links.used++]= (uint32_t)lin_page; // "master table"
 }
@@ -1050,7 +1134,11 @@ static void PAGING_LinkPageNew(Bitu lin_page, Bitu phys_page, Bitu linkmode, boo
 void PAGING_LinkPage(Bitu lin_page,Bitu phys_page) {
 	PageHandler * handler=MEM_GetPageHandler(phys_page);
 	Bitu lin_base=lin_page << 12;
-	if (lin_page>=TLB_SIZE || phys_page>=TLB_SIZE)
+
+	//NTS: phys_page is not used to index anything, however it must not use bits 31-30 used for other info. Stay within PHYSPAGE_ADDR.
+	phys_page &= PHYSPAGE_ADDR;
+
+	if (lin_page>=TLB_SIZE)
 		return E_Exit("Illegal page");
 
 	if (paging.links.used>=PAGING_LINKS) {
@@ -1093,7 +1181,7 @@ void PAGING_SwitchCPL(bool isUser) {
 			Bitu tlb_index = paging.krw_links.entries[i];
 			Bitu phys_page = paging.tlb.phys_page[tlb_index];
 			Bitu lin_base = tlb_index << 12;
-			bool dirty = (phys_page & PHYSPAGE_DITRY)? true:false;
+			bool dirty = (phys_page & PHYSPAGE_DIRTY)? true:false;
 			phys_page &= PHYSPAGE_ADDR;
 			PageHandler* handler = MEM_GetPageHandler(phys_page);
 			
@@ -1154,7 +1242,7 @@ void PAGING_SwitchCPL(bool isUser) {
 			for(Bitu i = 0; i < paging.ur_links.used; i++) {
 				Bitu tlb_index = paging.ur_links.entries[i];
 				Bitu phys_page = paging.tlb.phys_page[tlb_index];
-				bool dirty = (phys_page & PHYSPAGE_DITRY)? true:false;
+				bool dirty = (phys_page & PHYSPAGE_DIRTY)? true:false;
 				phys_page &= PHYSPAGE_ADDR;
 				PageHandler* handler = MEM_GetPageHandler(phys_page);
 
@@ -1234,7 +1322,11 @@ void PAGING_MapPage(Bitu lin_page,Bitu phys_page) {
 void PAGING_LinkPage(Bitu lin_page,Bitu phys_page) {
 	PageHandler * handler=MEM_GetPageHandler(phys_page);
 	Bitu lin_base=lin_page << 12;
-	if (lin_page>=(TLB_SIZE*(TLB_BANKS+1)) || phys_page>=(TLB_SIZE*(TLB_BANKS+1))) 
+
+	//NTS: phys_page is not used to index anything, however it must not use bits 31-30 used for other info. Stay within PHYSPAGE_ADDR.
+	phys_page &= PHYSPAGE_ADDR;
+
+	if (lin_page>=(TLB_SIZE*(TLB_BANKS+1)))
 		E_Exit("Illegal page");
 
 	if (paging.links.used>=PAGING_LINKS) {
@@ -1257,7 +1349,11 @@ void PAGING_LinkPage(Bitu lin_page,Bitu phys_page) {
 void PAGING_LinkPage_ReadOnly(Bitu lin_page,Bitu phys_page) {
 	PageHandler * handler=MEM_GetPageHandler(phys_page);
 	Bitu lin_base=lin_page << 12;
-	if (lin_page>=(TLB_SIZE*(TLB_BANKS+1)) || phys_page>=(TLB_SIZE*(TLB_BANKS+1))) 
+
+	//NTS: phys_page is not used to index anything, however it must not use bits 31-30 used for other info. Stay within PHYSPAGE_ADDR.
+	phys_page &= PHYSPAGE_ADDR;
+
+	if (lin_page>=(TLB_SIZE*(TLB_BANKS+1)))
 		E_Exit("Illegal page");
 
 	if (paging.links.used>=PAGING_LINKS) {
@@ -1349,12 +1445,19 @@ public:
 				table.block.a=1;		//Set access
 				phys_writed((PhysPt)((paging.base.page<<12)+(lin_page >> 10)*4),table.load);
 			}
-			if ((!entry.block.a) || (!entry.block.d)) {
-				entry.block.a=1;	//Set access
-				entry.block.d=1;	//Set dirty
-				phys_writed((table.block.base<<12)+(lin_page & 0x3ff)*4,entry.load);
+			if (do_pse && table.dirblock.ps) { // 4MB PSE page
+				phys_page =	 (table.dirblock4mb.base22<<10u)|
+						((table.dirblock4mb.base32&enable_pse_extmask)<<20u)|
+						 (lin_page&0x3FFu);
+			}
+			else {
+				if ((!entry.block.a) || (!entry.block.d)) {
+					entry.block.a=1;	//Set access
+					entry.block.d=1;	//Set dirty
+					phys_writed((table.block.base<<12)+(lin_page & 0x3ff)*4,entry.load);
+				}
+				phys_page=entry.block.base;
 			}
-			phys_page=entry.block.base;
 			PAGING_LinkPage(lin_page,phys_page);
 		} else {
 			if (lin_page<LINK_START) phys_page=paging.firstmb[lin_page];
@@ -1402,11 +1505,18 @@ public:
 				table.block.a=1;		//Set access
 				phys_writed((PhysPt)((paging.base.page<<12)+(lin_page >> 10)*4),table.load);
 			}
-			if (!entry.block.a) {
-				entry.block.a=1;	//Set access
-				phys_writed((table.block.base<<12)+(lin_page & 0x3ff)*4,entry.load);
+			if (do_pse && table.dirblock.ps) { // 4MB PSE page
+				phys_page =	 (table.dirblock4mb.base22<<10u)|
+						((table.dirblock4mb.base32&enable_pse_extmask)<<20u)|
+						 (lin_page&0x3FFu);
+			}
+			else {
+				if (!entry.block.a) {
+					entry.block.a=1;	//Set access
+					phys_writed((table.block.base<<12)+(lin_page & 0x3ff)*4,entry.load);
+				}
+				phys_page=entry.block.base;
 			}
-			phys_page=entry.block.base;
 		} else {
 			if (lin_page<LINK_START) phys_page=paging.firstmb[lin_page];
 			else phys_page=lin_page;
@@ -1537,3 +1647,56 @@ void POD_Load_CPU_Paging( std::istream& stream )
 		paging.tlb.writehandler[lcv] = &init_page_handler;
 	}
 }
+
+uint8_t PageHandler_HostPtReadB(PageHandler *p,PhysPt addr) {
+	return p->GetHostReadPt(PAGING_GetPhysicalPageNumber(addr))[addr&0xFFF];
+}
+
+uint16_t PageHandler_HostPtReadW(PageHandler *p,PhysPt addr) {
+	if ((addr&0xFFF) < 0xFFF) {
+		return host_readw(&(p->GetHostReadPt(PAGING_GetPhysicalPageNumber(addr))[addr&0xFFF]));
+	}
+	else {
+		return 	 PageHandler_HostPtReadB(p,addr) +
+			(PageHandler_HostPtReadB(p,addr+PhysPt(1u)) << 8u);
+	}
+}
+
+uint32_t PageHandler_HostPtReadD(PageHandler *p,PhysPt addr) {
+	if ((addr&0xFFF) < 0xFFD) {
+		return host_readd(&(p->GetHostReadPt(PAGING_GetPhysicalPageNumber(addr))[addr&0xFFF]));
+	}
+	else {
+		return 	 PageHandler_HostPtReadB(p,addr) +
+			(PageHandler_HostPtReadB(p,addr+PhysPt(1u)) << 8u) +
+			(PageHandler_HostPtReadB(p,addr+PhysPt(2u)) << 16u) +
+			(PageHandler_HostPtReadB(p,addr+PhysPt(3u)) << 24u);
+	}
+}
+
+void PageHandler_HostPtWriteB(PageHandler *p,PhysPt addr,uint8_t val) {
+	p->GetHostWritePt(PAGING_GetPhysicalPageNumber(addr))[addr&0xFFF] = val;
+}
+
+void PageHandler_HostPtWriteW(PageHandler *p,PhysPt addr,uint16_t val) {
+	if ((addr&0xFFF) < 0xFFF) {
+		host_writew(&(p->GetHostWritePt(PAGING_GetPhysicalPageNumber(addr))[addr&0xFFF]),val);
+	}
+	else {
+		PageHandler_HostPtWriteB(p,addr,            val     &0xFFu);
+		PageHandler_HostPtWriteB(p,addr+PhysPt(1u),(val>>8u)&0xFFu);
+	}
+}
+
+void PageHandler_HostPtWriteD(PageHandler *p,PhysPt addr,uint32_t val) {
+	if ((addr&0xFFF) < 0xFFD) {
+		host_writed(&(p->GetHostWritePt(PAGING_GetPhysicalPageNumber(addr))[addr&0xFFF]),val);
+	}
+	else {
+		PageHandler_HostPtWriteB(p,addr,            val      &0xFFu);
+		PageHandler_HostPtWriteB(p,addr+PhysPt(1u),(val>> 8u)&0xFFu);
+		PageHandler_HostPtWriteB(p,addr+PhysPt(2u),(val>>16u)&0xFFu);
+		PageHandler_HostPtWriteB(p,addr+PhysPt(3u),(val>>24u)&0xFFu);
+	}
+}
+
